# BSD 3-Clause License
#
# Copyright (c) 2024, Matheus Gomes Cordeiro 
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from mpmath import mp, matrix

def psi_n_single_fock_single_position(n, x, prec):
    """
    Calculates the nth wavefunction for a real scalar `x` with arbitrary precision using mpmath.

    Args:
        n (int): Quantum state number.
        x (float): Position at which to evaluate the wavefunction.
        prec (numpy.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.ctx_mp_python.mpf` : evaluated wavefunction with the specified precision.

    Examples:
        >>> psi_n_single_fock_single_position(0, 1.0, 60)
        mpf('0.45558067201133253483370525689785138607662639040929439687915331')
        >>> psi_n_single_fock_single_position(61, 1.0, 60)
        mpf('-0.239304919917113097789996116536717211865611421191819349290628243')

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision 
        floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """

    mp.dps = prec
    x = mp.mpf(str(x))
    n = mp.mpf(str(n))
    return  mp.hermite(n, x) * mp.power(mp.mpf('2.0'),(mp.mpf('-0.5') * n)) * mp.power(mp.gamma(n+1.0), mp.mpf('-0.5')) * mp.power(mp.pi , mp.mpf('-0.25')) * mp.exp(-mp.power(x , mp.mpf('2.0')) / mp.mpf('2.0')) 

def psi_n_single_fock_single_position_complex(n, x, prec):
    """
    Calculates the nth wavefunction for a complex scalar `x` with arbitrary precision using mpmath.

    Args:
        n (np.uint64): Quantum state number.
        x (np.complex128): Position at which to evaluate the wavefunction.
        prec (np.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.ctx_mp_python.mpc` : The evaluated wavefunction with the specified precision.

    Examples:
        >>> psi_n_single_fock_single_position_complex(0, 1.0 + 2.0j, 60)
        mpc(real='-1.40087973302624535996319358379185603705205815719366827159881527', imag='-3.06097806029750039193292973729038840279841978760336147713769087')
        >>> psi_n_single_fock_single_position_complex(61, 1.0 + 2.0j, 60)
        mpc(real='-511062135.475553070892329856229109412939170026007243421420322129', imag='131445997.757536932748911867174534983962121585813389430606204944')

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """

    mp.dps = prec
    x = mp.mpc(str(x.real),str(x.imag))
    n = mp.mpf(str(n))
    return  mp.hermite(n, x) * mp.power(mp.mpf('2.0'),(mp.mpf('-0.5') * n)) * mp.power(mp.gamma(n+1.0), mp.mpf('-0.5')) * mp.power(mp.pi , mp.mpf('-0.25')) * mp.exp(-mp.power(x , mp.mpf('2.0')) / mp.mpf('2.0')) 

def psi_n_single_fock_multiple_position(n, X, prec):
    """
    Calculates the nth wavefunction for a real vector `x` with arbitrary precision using mpmath.

    Args:
        n (np.uint64): Quantum state number.
        x (np.ndarray[np.float64]): Positions at which to evaluate the wavefunction.
        prec (np.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.matrices.matrices._matrix` : The evaluated wavefunction with the specified precision.

    Examples:
        >>> psi_n_single_fock_multiple_position(0, np.array([1.0, 2.0]), 20)
        matrix(
        [['0.45558067201133253483', '0.10165378830641791152']])
        >>> psi_n_single_fock_multiple_position(61, np.array([1.0, 2.0]), 20)
        matrix(
        [['-0.23930491991711309779', '-0.016773782204892582343']])

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """

    return  matrix([psi_n_single_fock_single_position(n, x, prec) for x in X]).T

def psi_n_single_fock_multiple_position_complex(n, X, prec):
    """
    Calculates the nth wavefunction for a complex vector `x` with arbitrary precision using mpmath.

    Args:
        n (np.uint64): Quantum state number.
        x (np.ndarray[np.complex128]): Positions at which to evaluate the wavefunction.
        prec (np.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.matrices.matrices._matrix` : The evaluated wavefunction with the specified precision.

    Examples:
        >>> psi_n_single_fock_multiple_position_complex(0, np.array([1.0 + 1.0j, 2.0 + 2.0j]), 20)
        matrix(
        [[mpc(real='0.40583486367087033308603', imag='-0.63205035161528260798606'), 
        mpc(real='-0.49096842060721693717778', imag='0.56845368634059468652777')]])
        >>> psi_n_single_fock_multiple_position_complex(61, np.array([1.0 + 1.0j, 2.0 + 2.0j]), 20)
        matrix(
        [[mpc(real='-7565.4894098859360141926', imag='921.4986211518276840917'), 
        mpc(real='-164189541.53192908120809', imag='-370892077.23796911662203')]])

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """


    return  matrix([psi_n_single_fock_single_position_complex(n, x, prec) for x in X]).T

def psi_n_multiple_fock_single_position(n, x, prec):

    """
    Determines the wavefunction for a real scalar `x` for all Fock states up to `n`, employing mpmath for arbitrary-precision calculations.

    Args:
        n (np.uint64): Quantum state number up to which the wavefunctions are calculated.
        x (np.float64): Position at which to evaluate the wavefunctions.
        prec (np.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.matrices.matrices._matrix` : The evaluated wavefunctions with the specified precision.

    Examples:
        >>> psi_n_multiple_fock_single_position(1, 1.0, 60)
        matrix(
        [['0.455580672011332534833705256897851386076626390409294396879153', 
        '0.644288365113475181510837645362740498634994248687269122618738']])

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """

    return  matrix([psi_n_single_fock_single_position(i, x, prec) for i in range(n+1)]).T

def psi_n_multiple_fock_single_position_complex(n, x, prec):
    """
    Determines the wavefunction for a complex scalar `x` for all Fock states up to `n`, employing mpmath for arbitrary-precision calculations.

    Args:
        n (np.uint64): Quantum state number up to which the wavefunctions are calculated.
        x (np.complex128): Position at which to evaluate the wavefunctions.
        prec (np.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.matrices.matrices._matrix` : The evaluated wavefunctions with the specified precision.

    Examples:
        >>> c_wavefunction_mmod_arb_prec(1, 1.0 + 2.0j, 20)
        matrix(
        [[mpc(real='-1.400879733026245359964', imag='-3.0609780602975003919354'), 
        mpc(real='6.6766102562991123531695', imag='-8.2911629223978481324862')]])

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """

    return  matrix([psi_n_single_fock_single_position_complex(i, x, prec) for i in range(n+1)]).T

def psi_n_multiple_fock_multiple_position(n, X, prec):   
    """
    Determines the wavefunction for a real vector `x` for all Fock states up to `n`, employing mpmath for arbitrary-precision calculations.

    Args:
        n (np.uint64): Quantum state number up to which the wavefunctions are calculated.
        x (np.ndarray[np.float64]): Positions at which to evaluate the wavefunctions.
        prec (np.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.matrices.matrices._matrix` : The evaluated wavefunctions with the specified precision.

    Examples:
        >>> wavefunction_mmmd_arb_prec(1, np.array([1.0, 2.0]), 20)
        matrix(
        [['0.45558067201133253483', '0.10165378830641791152'],
        ['0.64428836511347518151', '0.28752033217907949445']])

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """

    return  matrix([[psi_n_single_fock_single_position(i, x, prec) for x in X] for i in range(n+1)])

def psi_n_multiple_fock_multiple_position_complex(n, X, prec):   
    """
    Determines the wavefunction for a complex vector `x` for all Fock states up to `n`, employing mpmath for arbitrary-precision calculations.

    Args:
        n (np.uint64): Quantum state number up to which the wavefunctions are calculated.
        x (np.ndarray[np.complex128]): Positions at which to evaluate the wavefunctions.
        prec (np.uint64): Desired precision for the calculation (number of decimal digits).

    Returns:
        `mpmath.matrices.matrices._matrix` : The evaluated wavefunctions with the specified precision.

    Examples:
        >>> c_wavefunction_mmmd_arb_prec(1, np.array([1.0 + 1.0j, 2.0 + 2.0j]), 20)
        [[mpc(real='0.40583486367087033308603', imag='-0.63205035161528260798606'), 
        mpc(real='-0.49096842060721693717778', imag='0.56845368634059468652777')],
        [mpc(real='1.4677913476441970351171', imag='-0.31991701106983521979673'), 
        mpc(real='-2.9964982238469495343176', imag='0.21916142736845211639935')]]

    References:
        1. The mpmath development team. (2023). mpmath: a Python library for arbitrary-precision floating-point arithmetic (version 1.3.0). Retrieved from http://mpmath.org/
    """

    return  matrix([[psi_n_single_fock_single_position_complex(i, x, prec) for x in X] for i in range(n+1)])



import pytest
import numpy as np
from sympy import symbols
import src.fast_wave.wavefunction_numba as wn

@pytest.fixture(scope="module", autouse=True)
def initialize_c_s_matrix():
    """
    Fixture to initialize the global variable c_s_matrix before running tests.
    """
    global c_s_matrix
    c_s_matrix = wn.create_normalized_hermite_coefficients_matrix(60)

def test_hermite_sympy():
    """
    Tests the hermite_sympy function to verify the accuracy of Hermite polynomial computation.
    """
    x =  symbols("x")
    h0 = wn.hermite_sympy(0)
    h1 = wn.hermite_sympy(1)
    h2 = wn.hermite_sympy(2)

    assert h0 == 1
    assert h1 == 2 * x
    assert h2 == 4 * x**2 - 2

def test_create_hermite_coefficients_table():
    """
    Tests the create_normalized_hermite_coefficients_table function to verify if the normalized coefficient matrix is correct.
    """
    n_max = 2
    coeffs_table = wn.create_normalized_hermite_coefficients_matrix(n_max)
    
    expected_table = np.zeros((3, 3))
    expected_table[0, 2] = 0.75112554 
    expected_table[1, 1] = 1.06225193  
    expected_table[2, 0] = 1.06225193 
    expected_table[2, 2] = -0.53112597

    assert np.allclose(coeffs_table, expected_table)

def test_wavefunction_computation():
    """
    Tests the basic functionality of all wavefunction functions.
    """

    # Testing basic functionality
    test_output_sfsp_CS_matrix_2 = wn.psi_n_single_fock_single_position(2, 10.0)
    assert isinstance(test_output_sfsp_CS_matrix_2, float)

    test_output_sfsp_61 = wn.psi_n_single_fock_single_position(61, 10.0)
    assert isinstance(test_output_sfsp_61, float)

    test_output_sfsp_2 = wn.psi_n_single_fock_single_position(2, 10.0, CS_matrix = False)
    assert isinstance(test_output_sfsp_2, float)

    test_output_sfsp_61 = wn.psi_n_single_fock_single_position(61, 10.0, CS_matrix = False)
    assert isinstance(test_output_sfsp_61, float)
    
    test_output_mfsp = wn.psi_n_multiple_fock_single_position(2, 10.0)
    assert isinstance(test_output_mfsp , np.ndarray)
    
    test_output_sfmp_CS_matrix_2 = wn.psi_n_single_fock_multiple_position(2, np.array([10.0, 4.5]))
    assert isinstance(test_output_sfmp_CS_matrix_2, np.ndarray)

    test_output_sfmp_61 = wn.psi_n_single_fock_multiple_position(61, np.array([10.0, 4.5]))
    assert isinstance(test_output_sfmp_61, np.ndarray)

    test_output_sfmp_2 = wn.psi_n_single_fock_multiple_position(2, np.array([10.0, 4.5]), CS_matrix = False)
    assert isinstance( test_output_sfmp_2, np.ndarray)

    test_output_sfmp_61 = wn.psi_n_single_fock_multiple_position(61, np.array([10.0, 4.5]), CS_matrix = False)
    assert isinstance(test_output_sfmp_61, np.ndarray)
    
    test_output_mfmp = wn.psi_n_multiple_fock_multiple_position(2, np.array([10.0, 4.5]))
    assert isinstance(test_output_mfmp, np.ndarray)
    
    test_output_sfsp_c_CS_matrix_2 = wn.psi_n_single_fock_single_position_complex(2, 10.0 + 0.0j)
    assert isinstance(test_output_sfsp_c_CS_matrix_2, complex)

    test_output_sfsp_c_61 = wn.psi_n_single_fock_single_position_complex(61, 10.0 + 0.0j)
    assert isinstance(test_output_sfsp_c_61, complex)

    test_output_sfsp_c_2 = wn.psi_n_single_fock_single_position_complex(2, 10.0 + 0.0j, CS_matrix = False)
    assert isinstance(test_output_sfsp_c_2, complex)

    test_output_sfsp_c_61 = wn.psi_n_single_fock_single_position_complex(61, 10.0 + 0.0j, CS_matrix = False)
    assert isinstance(test_output_sfsp_c_61, complex)
    
    test_output_mfsp_c = wn.psi_n_multiple_fock_single_position_complex(2, 10.0 + 0.0j)
    assert isinstance(test_output_mfsp_c, np.ndarray)
    
    test_output_sfmp_c_CS_matrix_2 = wn.psi_n_single_fock_multiple_position_complex(2, np.array([10.0 + 0.0j, 4.5 + 0.0j]))
    assert isinstance(test_output_sfmp_c_CS_matrix_2, np.ndarray)

    test_output_sfmp_c_61 = wn.psi_n_single_fock_multiple_position_complex(61, np.array([10.0 + 0.0j, 4.5 + 0.0j]))
    assert isinstance(test_output_sfmp_c_61, np.ndarray)

    test_output_sfmp_c_2 = wn.psi_n_single_fock_multiple_position_complex(2, np.array([10.0 + 0.0j, 4.5 + 0.0j]), CS_matrix = False)
    assert isinstance(test_output_sfmp_c_2, np.ndarray)

    test_output_sfmp_c_61 = wn.psi_n_single_fock_multiple_position_complex(61, np.array([10.0 + 0.0j, 4.5 + 0.0j]), CS_matrix = False)
    assert isinstance(test_output_sfmp_c_61, np.ndarray)
    
    test_output_mfmp_c = wn.psi_n_multiple_fock_multiple_position_complex(2, np.array([10.0 + 0.0j, 4.5 + 0.0j]))
    assert isinstance(test_output_mfmp_c, np.ndarray)
    
    print("All functionality tests passed.")




# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import lru_cache, wraps
import numpy as np

def int_array_cache_Numba_single_fock(fn):
    """
    Cache decorator for functions that receive real multiple positions (numpy array) 
    and is a problem Single Fock in the Numba module.

    This decorator caches function results to improve performance, particularly when `fn` 
    is called multiple times with the same arguments. The function to be decorated must accept 
    an integer `n`, a numpy array `x_array`, and a boolean `CS_matrix`. The numpy array is 
    converted to a tuple for caching purposes, as `lru_cache` only accepts hashable types.

    Args:
        fn (callable): The function to be decorated, which takes three arguments:
            - n (np.uint64): The state number.
            - x_array (np.ndarray[np.float64]): Array representing the positions.
            - CS_matrix (bool): A flag to indicate CS matrix.

    Returns:
        `callable` : A wrapped version of **fn** with caching enabled, including methods to access cache information.

    References:
        1. This code is a modified version of the `tensor_int_cache` provided in Mr Mustard ([link](https://github.com/XanaduAI/MrMustard/blob/develop/mrmustard/math/caching.py#L26)),
        which is released under Apache License, Version 2.0, with the following copyright notice: Copyright 2022 Xanadu Quantum Technologies Inc. All rights reserved.
    """
    @lru_cache
    def cached_wrapper(n, x_tuple, CS_matrix=True):
        x_array = np.array(x_tuple, dtype=np.float64)
        return fn(n, x_array, CS_matrix)

    @wraps(fn)
    def wrapper(n, x_array, CS_matrix=True):
        return cached_wrapper(n, tuple(x_array), CS_matrix)

    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper


def int_array_cache_Numba_multiple_fock(fn):
    """
    Cache decorator for functions that receive real multiple positions (numpy array) 
    and is a problem Multiple Fock in the Numba module.

    This decorator caches function results to improve performance, particularly when `fn` 
    is called multiple times with the same arguments. The function to be decorated must accept 
    an integer `n`, a numpy array `x_array`, and a boolean `CS_matrix`. The numpy array is 
    converted to a tuple for caching purposes, as `lru_cache` only accepts hashable types.

    Args:
        fn (callable): The function to be decorated, which takes two arguments:
            - n (np.uint64): The state number.
            - x_array (np.ndarray[np.float64]): Array representing the positions.

    Returns:
        `callable` : A wrapped version of **fn** with caching enabled, including methods to access cache information.

    References:
        1. This code is a modified version of the `tensor_int_cache` provided in Mr Mustard ([link](https://github.com/XanaduAI/MrMustard/blob/develop/mrmustard/math/caching.py#L26)),
        which is released under Apache License, Version 2.0, with the following copyright notice: Copyright 2022 Xanadu Quantum Technologies Inc. All rights reserved.
    """
    @lru_cache
    def cached_wrapper(n, x_tuple):
        x_array = np.array(x_tuple, dtype=np.float64)
        return fn(n, x_array)

    @wraps(fn)
    def wrapper(n, x_array):
        return cached_wrapper(n, tuple(x_array))

    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper


def int_array_cache_Cython(fn):
    """
    Cache decorator for functions that receive real multiple positions (numpy array) 
    in the Cython module.

    This decorator caches function results to improve performance, particularly when `fn` 
    is called multiple times with the same arguments. The function to be decorated must accept 
    an integer `n` and a numpy array `x_array`. The numpy array is converted to a tuple for 
    caching purposes, as `lru_cache` only accepts hashable types.

    Args:
        fn (callable): The function to be decorated, which takes two arguments:
            - n (np.uint64): The state number.
            - x_array (np.ndarray[np.float64]): Array representing the positions.

    Returns:
        `callable` : A wrapped version of **fn** with caching enabled, including methods to access cache information.

    References:
        1. This code is a modified version of the `tensor_int_cache` provided in Mr Mustard ([link](https://github.com/XanaduAI/MrMustard/blob/develop/mrmustard/math/caching.py#L26)),
        which is released under Apache License, Version 2.0, with the following copyright notice: Copyright 2022 Xanadu Quantum Technologies Inc. All rights reserved.
    """
    @lru_cache
    def cached_wrapper(n, x_tuple):
        x_array = np.array(x_tuple, dtype=np.float64)
        return fn(n, x_array)

    @wraps(fn)
    def wrapper(n, x_array):
        return cached_wrapper(n, tuple(x_array))

    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper


def int_array_cache_Numba_complex_single_fock(fn):
    """
    Cache decorator for functions that receive complex multiple positions (numpy array) 
    and is a problem Single Fock in the Numba module.

    This decorator caches function results to improve performance, particularly when `fn` 
    is called multiple times with the same arguments. The function to be decorated must accept 
    an integer `n`, a numpy array `x_array`, and a boolean `CS_matrix`. The numpy array is 
    converted to a tuple for caching purposes, as `lru_cache` only accepts hashable types.

    Args:
        fn (callable): The function to be decorated, which takes three arguments:
            - n (np.uint64): The state number.
            - x_array (np.ndarray[np.complex128]): Array representing the positions.
            - CS_matrix (bool): A flag to indicate CS matrix.

    Returns:
        `callable` : A wrapped version of **fn** with caching enabled, including methods to access cache information.
         
    References:
        1. This code is a modified version of the `tensor_int_cache` provided in Mr Mustard ([link](https://github.com/XanaduAI/MrMustard/blob/develop/mrmustard/math/caching.py#L26)),
        which is released under Apache License, Version 2.0, with the following copyright notice: Copyright 2022 Xanadu Quantum Technologies Inc. All rights reserved.
    """
    @lru_cache
    def cached_wrapper(n, x_tuple, CS_matrix):
        x_array = np.array(x_tuple, dtype=np.complex128)
        return fn(n, x_array, CS_matrix)

    @wraps(fn)
    def wrapper(n, x_array, CS_matrix):
        return cached_wrapper(n, tuple(x_array), CS_matrix)

    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper


def int_array_cache_Numba_complex_multiple_fock(fn):
    """
    Cache decorator for functions that receive complex multiple positions (numpy array) 
    in the Numba module.

    This decorator caches function results to improve performance, particularly when `fn` 
    is called multiple times with the same arguments. The function to be decorated must accept 
    an integer `n`, a numpy array `x_array`, and a boolean `CS_matrix`. The numpy array is 
    converted to a tuple for caching purposes, as `lru_cache` only accepts hashable types.

    Args:
        fn (callable): The function to be decorated, which takes two arguments:
            - n (np.uint64): The state number.
            - x_array (np.ndarray[np.complex128]): Array representing the positions.

    Returns:
        `callable` : A wrapped version of **fn** with caching enabled, including methods to access cache information.

    References:
        1. This code is a modified version of the `tensor_int_cache` provided in Mr Mustard ([link](https://github.com/XanaduAI/MrMustard/blob/develop/mrmustard/math/caching.py#L26)),
        which is released under Apache License, Version 2.0, with the following copyright notice: Copyright 2022 Xanadu Quantum Technologies Inc. All rights reserved.
    """
    @lru_cache
    def cached_wrapper(n, x_tuple):
        x_array = np.array(x_tuple, dtype=np.complex128)
        return fn(n, x_array)

    @wraps(fn)
    def wrapper(n, x_array):
        return cached_wrapper(n, tuple(x_array))

    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper

def int_array_cache_Cython_complex(fn):
    """
    Cache decorator for functions that receive complex multiple positions (numpy array)
    and is a problem Single Fock in the Cython module.

    This decorator caches function results to improve performance, particularly when `fn`
    is called multiple times with the same arguments. The function to be decorated must accept 
    an integer `n` and a numpy array `x_array`. The numpy array is converted to a tuple for 
    caching purposes, as `lru_cache` only accepts hashable types.

    Args:
        fn (callable): The function to be decorated. It must take two arguments:
            - n (np.uint64): The state number.
            - x_array (np.ndarray of dtype=np.complex128): The positions.

    Returns:
        `callable` : A wrapped version of **fn** with caching enabled. This wrapped function also exposes the following methods for cache management.

    References:
        1. This code is a modified version of the `tensor_int_cache` provided in Mr Mustard ([link](https://github.com/XanaduAI/MrMustard/blob/develop/mrmustard/math/caching.py#L26)),
        which is released under Apache License, Version 2.0, with the following copyright notice: Copyright 2022 Xanadu Quantum Technologies Inc. All rights reserved.
    """
    @lru_cache
    def cached_wrapper(n, x_tuple):
        x_array = np.array(x_tuple, dtype=np.complex128)
        return fn(n, x_array)

    @wraps(fn)
    def wrapper(n, x_array):
        return cached_wrapper(n, tuple(x_array))

    wrapper.cache_info = cached_wrapper.cache_info
    wrapper.cache_clear = cached_wrapper.cache_clear

    return wrapper



"""
    This project presents an optimized approach for calculating the position wave functions of a Fock state of a quantum harmonic oscillator, with applications in Photonic Quantum Computing simulations. Leveraging [Numba](https://numba.pydata.org/)  [[1](#references)] and [Cython](https://cython.org/) [[2](#references)], this approach outperforms the [Mr Mustard](https://mrmustard.readthedocs.io/en/stable/) package [[3, 4](#references)] in computing a single wave function value at a single position and at multiple positions.

    ## ✨ Advantages


    - **Highly Efficient**: This package includes two fixed-point modules focused on speed. One is implemented using *Numba*, an open-source Just-in-Time (JIT) compiler, and the other module is implemented in *Cython*, a programming language that combines the ease of use of Python with the speed of C.
    - **Highly Accurate**: The functions in this package have precision next to the precision of Wolfram Mathematica and MATLAB. In addition, there is a module for calculating wave functions with arbitrary precision using the *mpmath* package.
    - **Past response cache**: This package provides a caching module designed to enhance the performance of functions that take multiple positions of a *NumPy* array as input. This module stores previously computed results by leveraging Python's functools.lru_cache, eliminating the need for redundant calculations. This caching module is inspired by the [caching module](https://github.com/XanaduAI/MrMustard/blob/develop/mrmustard/math/caching.py#L26) from Mr. Mustard, a package from the photonic quantum computing company Xanadu.


    ## 🛠️ Setup
    To use this package, simply run the following command in the command line: 
    ```bash
    pip install fast-wave
    ``` 

    ## 🎨 Examples

    The functions `psi_n_multiple_fock_multiple_position` calculate the values of the wavefunction $\psi_{0 → n} \( X_m \)$ for multiple Fock states ($n$) and multiple positions ($X_{m}$).

    **Inputs:**

    * `n`: An integer specifying the maximum Fock state ($n$).
    * `X_m`: A 1D `numpy` array with `m` positions, representing the input values where $\psi_{0 → n}$ will be evaluated. For example, `np.array([1.0, 2.0])` has dimension $(m,)$, where $m=2$ in this case.

    **Outputs:**

    - `numpy` **Matrix**: The output has dimensions $(n+1) × m$, where:
        - $n+1$: Corresponds to the Fock states $[0, 1, ..., n]$.
        - $m$: Represents the positions given in `X_m`.

    **Demonstration:**

    Using the provided inputs:

    ```python
    >>> import fast_wave.wavefunction_numba as wn
    Functionality Test Passed: True
    >>> import fast_wave.wavefunction_cython as wc
    >>> import numpy as np
    >>> wn.psi_n_multiple_fock_multiple_position(1,np.array([1.0 ,2.0])) 
    array([[0.45558067, 0.10165379],
        [0.64428837, 0.28752033]])
    >>> wc.psi_n_multiple_fock_multiple_position(1,np.array([1.0 ,2.0]))
    array([[0.45558067, 0.10165379],
        [0.64428837, 0.28752033]])
    ```

    **Explanation of the Output:**

    * For `n=1` and `X_m = np.array([1.0, 2.0])`:
    - The output matrix has dimensions $(n+1) × m = 2 × 2$.
    - The first row contains $\psi_{0}(x_1)$ and $\psi_{0}(x_2)$.
    - The second row contains $\psi_{1}(x_1)$ and $\psi_{1}(x_2)$.

    There are other examples in the examples folder: [Speed Tests: Numba & Cython](https://colab.research.google.com/github/fobos123deimos/fast-wave/blob/main/examples/speed_tests_numba_and_cython.ipynb); [Precision Tests: mpmath](https://colab.research.google.com/github/fobos123deimos/fast-wave/blob/main/examples/precision_tests_mpmath.ipynb). In the first one there is a comparison with the [Mr Mustard](https://mrmustard.readthedocs.io/en/stable/) package.

    ## 🌊 The Wavefunction

    The wavefunction, $\Psi(y,t)$, is a fundamental concept in quantum mechanics that describes the quantum state of a particle or system. Its absolute square, $|\Psi(y,t)|^2$, represents the probability density of finding the particle at position $\mathbf{y}$ and time $\mathbf{t}$. Due to the normalization property: $\int_{-\infty}^{\infty} |\Psi(y,t)|^2 dy = 1$ it's guaranteed that for a given time $\mathbf{t}$, the total probability of finding the particle somewhere in space is unity [[5](#references)].

    ###  Schrödinger Equation

    The wavefunction is the solution to the Schrödinger equation, a fundamental equation in quantum mechanics:

    $$
    -\Bigg(\{\hbar^{2} \over 2m\}\Bigg) \{\partial^2 \Psi(y,t) \over \partial y^{2}\} + \Bigg(\{m\omega^2 y^2 \over 2\}\Bigg) \Psi(y,t) = \mathbf{i}\hbar \, \{\partial \Psi(y,t) \over \partial t\} \quad \mathbf{(1)}
    $$

    where $\mathbf{\hbar}$ is the reduced Planck constant, $\mathbf{m}$ is the mass of the particle, and $\mathbf{\omega}$ is the angular frequency of the harmonic potential. The symbol $\mathbf{i}$ represents the imaginary unit. When seeking the solution to this equation, we separated the variables as follows: $\Psi(y,t) = \psi(y)f(t)$, and we find as a result for $f(t)$ [[5](#references)]:

    $$ f(t) = C \, e^{-iEt/\hbar} \quad \mathbf{(2)}$$

    where $\mathbf{C}$ may be considered an arbitrary complex constant and $\mathbf{E}$, the system separation constant can be interpreted as the system's energy. Substituting into the wavefunction we have [[5](#references)]:

    $$ \Psi(y,t) = C \, e^{-iEt/\hbar} \, \psi(y) \quad \mathbf{(3)}$$

    The term $e^{-iEt/\hbar}$ is called the **phase factor** of $\Psi(y,t)$. In order to find $\psi(y)$ we then solve the **time-independent Schröndiger equation** [[5](#references)]:

    $$
    -\Bigg(\{\hbar^{2} \over 2m \}\Bigg) \psi''(y) + \Bigg(\{m\omega^2 y^2 \over 2 \}\Bigg) \psi(y) = E \, \psi(y) \quad \mathbf{(4)}
    $$


    ### Quantum Harmonic Oscillator

    By solving equation **(4)**, we obtain a family of energy eigenfunctions defined as follows [[5](#references)]:

    $$
    \psi_n(y) = \Bigg(\{ m\omega \over \pi\hbar \}\Bigg)^{1/4} \Bigg(\{ 1 \over \sqrt{2^n n!}\}\Bigg) H_n\Bigg(\sqrt{m\omega \over \hbar }y\Bigg) \, e^{-m\omega y^2/2\hbar} , \quad  n \in \mathbb{N}_{0} \quad \mathbf{(5)}
    $$

    where $\mathbf{n}$ represents a non-negative integer corresponding to the different energy states of the system, with energies given by $E_n = [n + (1/2)]\hbar \omega$. The term $H_n$ denotes the Hermite polynomial of degree $\mathbf{n}$; thus, for each energy state $\mathbf{n}$, there is an associated Hermite polynomial of degree $\mathbf{n}$ within its eigenfunction [[5](#references)]:

    <br>

    <div style="text-align: center;">
        <img src="https://github.com/user-attachments/assets/cd191cda-94cf-47f3-8835-85425e932b22" alt=" " style="max-width: 100%; ">
        <figcaption style="font-style: italic; color: #666; text-align: center;">
        <p style="font-style: italic; color: #666;"> Wavefunctions and energies for different $\mathbf{n}$ values. <a href="#references">[6]</a></p>
    </div>

    <br>

    The energy eigenfunction for an energy state $\mathbf{n}$ is the wavefunction for an energy state $\mathbf{n}$ of a Quantum Harmonic Oscillator. From this definition, we can then represent the wave function $\Psi(x,t)$ as a series expansion of its family of energy eigenfunctions $\{\psi_{n}(x)\}$ [[5](#references)]:

    $$
    \Psi(y,t) = \sum_{n=0}^{\infty} c_{n} \, \psi_{n}(y) \, e^{-\mathbf{i}E_{n}t/\hbar} \quad \mathbf{(6)}
    $$

    where $\mathbf{c_{n}}$ are complex constants that determine the contribution of each eigenfunction $\psi_{n}(y)$ to the total wavefunction $\Psi(y,t)$. These coefficients are chosen to ensure that the wavefunction satisfies the initial condition of the problem ($t=0$) [[5](#references)].

    ### Fock states

    When defining the dimensionless variable $x = \Big(m\omega/\hbar\Big)^{1/2}y$, referred to as the **reduced coordinate**, it follows that $dy = \Big(\hbar/m\omega\Big)^{1/2}dx$. As a result, we can write [[7](#references)]:


    $\displaystyle\int_{-\infty}^{+\infty} |\psi(y)|^{2} dy = 1 \implies \int_{-\infty}^{+\infty}\Bigg[\Bigg(\{m\omega \over \pi\hbar \}\Bigg)^{1/2} \Bigg(\{1 \over 2^n n!\}\Bigg) H_n^{2}\Bigg(\sqrt{ m\omega \over \hbar}y\Bigg) e^{-m\omega y^2/\hbar} \Bigg] dy \implies$ 


    $\displaystyle\int_{-\infty}^{+\infty} \Bigg[ \Bigg(\{m\omega \over \pi\hbar \} \Bigg)^{1/2} \Bigg( \{ 1 \over 2^n n! \} \Bigg) H_{n}^{2}(x) \, e^{-x^{2}}\Bigg]\Bigg[\{ \hbar \over \omega m \}\Bigg]^{1/2}dx = 1 \implies \int_{-\infty}^{+\infty} \Bigg[ \Bigg(\{ 1  \over 2^n n!\sqrt{\pi} \} \Bigg) H_{n}^{2}(x) \, e^{-x^{2}}\Bigg]dx = 1 \implies$


    $\displaystyle\int_{-\infty}^{+\infty} \, \Bigg| \Bigg( \{1 \over 2^n n!\sqrt{\pi} \} \Bigg)^{1/2} H_{n}(x) \, e^{-x^{2}/2} \Bigg|^{2}dx = 1 \implies \int_{-\infty}^{+\infty} |\psi(x)|^{2} dx = 1\implies$

    $$\psi_{n}(x) = \Bigg( \{1  \over 2^n n!\sqrt{\pi} \}\Bigg)^{1/2} H_{n}(x) \quad e^{-x^{2}/2} \quad \mathbf{(7)}$$

    This demonstrates that the wavefunction of a Quantum Harmonic Oscillator can be represented in a more dimensionless form, known as the **Hermite function**, it is also sometimes referred to as the **Gauss Hermite function** (equation $\mathbf{7}$) [[8](#references)]. A more simplified form for this type of function is this: $\psi_{n}(x) = \mathcal{N_{n}} \, H_{n}(x) \, e^{-x^{2}/2}$, where $\mathcal{N_{n}} = [1 /(2^n n!\sqrt{\pi})]^{1/2}$ is referred to as the normalization constant [[9](#references)].

    This type of representation for the wavefunction enables the modeling of wave functions for **Fock states**, which are energy eigenstates with a well-defined number of particles. When the particles are considered photons, $\mathbf{n}$, the degree of the Hermite polynomial, also represents the number of photons. Consequently, the Fock states ($\ket{n}$) become eigenstates of the photon number operator $\hat{n}$, such that $\hat{n}\ket{n}=n\ket{n}$. Moreover, the energy states now correspond to the energy levels of an optical system containing $n$-photons [[10](#references)].


    $\ket{\psi_{n}} = \displaystyle\{1 \over \sqrt{n!} \}(a^{\dagger})^{n}\ket{\psi_{0}} \implies \psi_{n}(q) = \{ 1 \over \sqrt{n!} \}\{ 1 \over \sqrt{2^n} \}\Bigg(q - \{\partial \over \partial q \}\Bigg)^{n}\psi_{0}(q) \implies $

    $\psi_{n}(q) = \displaystyle\{1 \over \sqrt{n!}\}\{1 \over \sqrt{2^n} \}  H_{n}(q)  \psi_{0}(q) \implies \psi_{n}(q) = \displaystyle\{1 \over \sqrt{n!} \}\{1  \over \sqrt{2^n} \}  H_{n}(q) \Big(\pi^{-1/4}\Big) e^{-q^{2}/2} \implies$

    $$\psi_{n}(q) = \displaystyle\Bigg(\{1 \over 2^n n!\sqrt{\pi} \}\Bigg)^{1/2} H_{n}(q) \quad e^{-q^{2}/2} \quad \mathbf{(8)}$$

    where $\mathbf{q}$ is interpreted as the normalized field amplitude associated with the position quadrature ($\hat{q}$​) and has a dimensionless character similar to $\mathbf{x}$ [[10](#references)]. 

    The wavefunction $\mathbf{n}$ of a Fock state is the projection of the state $\ket{n}$, and as we can observe, the wavefunctions of a Fock state for $n$-photon systems can be interpreted as a family of dimensionless energy eigenfunctions (Hermite functions), similar to those shown by equation $\mathbf{(7)}$ [[10](#references)].

    ## 🔁 The Wavefunction Recurrence

    In essence, Mr Mustard's strategy is to use the [Renormalized Hermite Polynomial](https://mrmustard.readthedocs.io/en/stable/code/api/mrmustard.math.hermite_renormalized.html) [[3, 4](#references)] for the computation of the wavefunction of a quantum harmonic oscillator. Below, we show the recurrence for calculating the Renormalized Hermite Polynomial, as well as the method for calculating it using the traditional Hermite polynomial:

    $$H_{n+1}^{re}(x) = \displaystyle\{2 \over \sqrt{n+1} \}\Big[xH_{n}^{re}(x) - H_{n-1}^{re}(x)\sqrt{n-1}\Big] \quad \mathbf{(8)} $$ 

    $$H_{n}^{re}(x) = \displaystyle\{H_{n}(x) \over \sqrt{n!}} \quad \mathbf{(10)} $$ 

    When we use this polynomial in calculating the wavefunction of a Quantum Harmonic Oscillator, the equation is as follows:

    $$\psi_{n}(x) = \displaystyle\Bigg(\{1 \over 2^n\sqrt{\pi} \}\Bigg)^{1/2}H_{n}^{re}(x) \quad e^{-\{x^2 \over 2 \}} \quad \mathbf{(11)} $$ 

    In this package, we implemented a recurrence based on the recursive solution to the wavefunction of the Quantum Harmonic Oscillator presented in the work of *José Maria Pérez-Jordá* [[11](#references)]. The recurrence we implemented was for $\psi_{n+1}$, which we obtained from the recursive definition of the Hermite polynomial [[12](#references)], as suggested by *José Maria Pérez-Jordá* in his article:


    $H_{n+1}(x) = 2xH_{n}(x) - 2nH_{n-1}(x) \implies $


    $\Bigg( \displaystyle\{e^{-x^2 / 2} \over \sqrt{2^{n-1}(n-1)!\pi^{1/2}}\}\Bigg)H_{n+1}(x) = \Bigg( \displaystyle\{e^{-x^2 / 2} \over \sqrt{2^{n-1}(n-1)!\pi^{1/2}}\}\Bigg)2xH_{n}(x) -\Bigg( \displaystyle\{e^{-x^2 / 2} \over \sqrt{2^{n-1}(n-1)!\pi^{1/2}}\}\Bigg)2nH_{n-1}(x) \implies$


    $\Bigg( \displaystyle\{e^{-x^{2}/2} \over \sqrt{2^{n-1}(n-1)!\pi^{1/2}}\}\Bigg)H_{n+1}(x) = \Bigg( \displaystyle\{e^{-x^{2}/2} \over \sqrt{2^{n-1}(n-1)!\pi^{1/2}}\}\Bigg)2xH_{n}(x) -2n\psi_{n-1}(x) \implies $


    $\displaystyle\{1 \over \sqrt{2n}\}\Bigg( \displaystyle\{e^{-x^2 / 2} \over \sqrt{2^{n-1}(n-1)!\pi^{1/2}}\}\Bigg)H_{n+1}(x) = \{1 \over \sqrt{2n} \}\Bigg( \displaystyle\{e^{-x^2 /2} \over \sqrt{2^{n-1}(n-1)!\pi^{1/2}}\}\Bigg)2xH_{n}(x) -\{2n \over \sqrt{2n}\} \, \psi_{n-1}(x) \implies$


    $\Bigg(\displaystyle\{e^{-x^2 / 2} \over \sqrt{2^n n!\pi^{1/2}}\}\Bigg) H_{n+1}(x) = 2x\psi_{n}(x) - \{2n \over \sqrt{2n}\} \, \psi_{n-1}(x) \implies$


    $\displaystyle\Bigg(\{1 \over \sqrt{2(n+1)} \}\Bigg)\Bigg(\displaystyle\{e^{-x^2 / 2}  \over \sqrt{2^n n!\pi^{1/2}} \}\Bigg) H_{n+1}(x) = \displaystyle\Bigg(\{1 \over \sqrt{2(n+1)}\}\Bigg)2x\psi_{n}(x) - \displaystyle\Bigg(\{1 \over \sqrt{2(n+1)}\}\Bigg) \, \{2n \over \sqrt{2n} \} \, \psi_{n-1}(x) \implies$


    $$\psi_{n+1}(x) = \displaystyle\Bigg(\sqrt{2 \over n+1}\Bigg)x\psi_{n}(x) -\Bigg(\sqrt{n \over n+1}\Bigg)\psi_{n-1}(x) \quad \mathbf{(12)}$$

    
    Besides the use of this recurrence in this package, the same authors implemented a version of it in a Cython module of QuTip: [_distributions.pyx](https://github.com/qutip/qutip/blob/master/qutip/_distributions.pyx), to be used in the HarmonicOscillatorWaveFunction class from the [distributions.py](https://github.com/qutip/qutip/blob/master/qutip/distributions.py) module.


    ## ⚡️The Numba Module - Hybrid Solution

    We use a hybrid solution with two algorithms for calculating the wave function for calculating a single Fock wave function's values at multiple positions (Single Fock and Multiple Position) (`psi_n_single_fock_multiple_position`). For $n>60$ or more than 35 positions, we use the recurrence for the wave function. For $n\le 60$ and at most 35 positions we use a precomputed matrix with the normalized coefficients of the Hermite polynomial as follows:


    $$\psi_{i}(x) = \displaystyle\{1 \over \sqrt{2^i i!\pi^{1/2}}\}H_{i}(x)e^{-x^{2}/2} = \{1 \over \sqrt{2^{i}i!\pi^{1/2}}\}\mathbf{C_{n}[i]} \cdot  \mathbf{x^{p}} e^{-x^{2}/2} \implies $$


    $$\psi_{i}(x) = \mathbf{C^{s}_{n}[i]\cdot x^{p}e^{-x^{2}/2} \quad \mathbf{(13)}}$$


    where $\mathbf{C^{s}_{n}[i]}$ is the row vector of normalized coefficients that multiply each power of $x$ up to $x^n$. The entire matrix $\mathbf{C^s_n}$ of such rows is precomputed up to degree $n=60$.  $\mathbf{x^{p}}$ is a column vector of powers up to n, with zeros in places where the coefficient is zero; for example, for $i=3$, $\mathbf{x^{p}} = [x^{3}, 0.0, x^{1}, 0.0]^T$. This hybrid algorithm is also used in Single Fock and Single Position (`psi_n_single_fock_single_position`) problems, though it offers no computational advantage in these cases. Additionally, there is an argument named **CS_matrix** for these Single Fock functions, set to **True** to enable the use of this matrix. In other words, you can use only the recurrence relation for the wave function at any value. The use of this coefficient matrix is limited to values up to **60** (determined empirically), as beyond this point, the function may encounter precision errors, resulting in incoherent outputs [[13](#references)].

    ## ⚡️ The Numba Module - Arguments

    For this algorithm to perform as efficiently as possible, Numba's Just-in-Time compilation is used in conjunction with [lru_cache (Least Recently Used - Cache Management)](https://docs.python.org/3/library/functools.html). The following arguments were used in the **@nb.jit** decorator:

    - **nopython=True:** This argument forces the Numba compiler to operate in "nopython" mode, which means that all the code within the function must be compilable to pure machine code without falling back to the Python interpreter. This results in significant performance improvements by eliminating the overhead of the Python interpreter.
    - **looplift=True:** This argument allows Numba to "lift" loops out of "nopython" mode. That is, if there are loops in the code that cannot be compiled in "nopython" mode, Numba will try to move them outside of the compiled part and execute them as normal Python code.
    - **nogil=True:** This argument releases the Python Global Interpreter Lock (GIL) while the function is executing. It is useful for allowing the Numba-compiled code to run in parallel with other Python threads, increasing performance in multi-threaded programs.
    - **boundscheck=False:** Disables array bounds checking. Normally, Numba checks if array indices are within valid bounds. Disabling this check can increase performance but may result in undefined behavior if there are out-of-bounds accesses.
    - **cache=True:** Enables caching of the compiled function. The first time the function is compiled, Numba stores the compiled version in a cache. On subsequent executions, Numba can reuse the compiled version from the cache instead of recompiling the function, reducing the function's startup time.

    ## ⚙️ The Cython Module

    The Cython module includes compiled files for Linux (**.so**) and Windows (**.pyd**), which allows it to be used in Google Colab (Linux). Additionally, this module supports three versions of Python 3: 3.10, 3.11, and 3.12. All these files are placed in the package folder upon installation. The source code of the Cython module is available in the repository in **.pyx** format. In the functions of the Cython module, some decorators are used to increase speed:

    - **@cython.nogil**: This decorator allows a Cython function to release the Global Interpreter Lock (GIL), making it possible to execute that block of code concurrently in multiple threads.
    - **@cython.cfunc**:  This decorator tells Cython to treat the function as a C function, meaning it can be called from other Cython or C code, not just Python. The function will have C-level calling conventions.
    - **@cython.locals(...)**: Declares local variable types to optimize performance.
    - **@cython.boundscheck(False)**: Disables bounds checking for arrays/lists to boost speed, but at the cost of safety. 


    ## 📖 References

    Our journey through the quantum realm is inspired by the following:

    1. Lam, S. K., Pitrou, A., & Seibert, S. (2015). _Numba: A LLVM-based Python JIT compiler_. In _Proceedings of the Second Workshop on the LLVM Compiler Infrastructure in HPC_ (LLVM '15) (pp. 7-12). Association for Computing Machinery. https://doi.org/10.1145/2833157.2833162
    2. Behnel, S., Bradshaw, R., Citro, C., Dalcin, L., Seljebotn, D. S., & Smith, K. (2011). *Cython: The best of both worlds*. Computing in Science & Engineering, 13(2), 31-39. https://doi.org/10.1109/MCSE.2010.118
    3. Yao, Y., Miatto, F., & Quesada, N. (2024). _Riemannian optimization of photonic quantum circuits in phase and Fock space_ [Preprint]. arXiv:2209.06069. [https://doi.org/10.21468/SciPostPhys.17.3.082](https://doi.org/10.21468/SciPostPhys.17.3.082)
    4. Miatto, F. M., & Quesada, N. (2020). *_Fast optimization of parametrized quantum optical circuits_* (*Quantum*, 4, 366). [https://doi.org/10.22331/q-2020-11-30-366](https://doi.org/10.22331/q-2020-11-30-366)
    5. Bowers, P. L. (2020). *Lectures on Quantum Mechanics: A Primer for Mathematicians*. Cambridge University Press. ISBN: [1108429769](https://www.worldcat.org/isbn/1108429769) ([9781108429764](https://www.worldcat.org/isbn/9781108429764))
    6. Aerts, D., Beltran, L. *Quantum Structure in Cognition: Human Language as a Boson Gas of Entangled Words*. Found Sci 25, 755–802 (2020). [https://doi.org/10.1007/s10699-019-09633-4](https://doi.org/10.1007/s10699-019-09633-4)
    7. Beiser, A. (2003). *Concepts of Modern Physics*. 6th ed. McGraw Hill. ISBN: [0072448482](https://www.worldcat.org/isbn/0072448482) ([9780072448481](https://www.worldcat.org/isbn/9780072448481))
    8. Celeghini, E., Gadella, M., & del Olmo, M. A. (2021). *Hermite functions and Fourier series*. Symmetry, 13(5), Article 853. [https://doi.org/10.3390/sym13050853](https://doi.org/10.3390/sym13050853)
    9. Schleich, W. P. (2001). *Quantum optics in phase space*. Wiley-VCH. ISBN: [352729435X](https://www.worldcat.org/isbn/352729435X) ([9783527294350](https://www.worldcat.org/isbn/9783527294350))
    10. Leonhardt, U. (2010). Essential Quantum Optics: From Quantum Measurements to Black Holes. Cambridge: Cambridge University Press. ISBN: [0521869781](https://www.worldcat.org/isbn/0521869781) ([9780521869782](https://www.worldcat.org/isbn/9780521869782))
    11. Pérez-Jordá, J. M. (2017). *On the recursive solution of the quantum harmonic oscillator*. European Journal of Physics, 39(1), 015402. [https://doi.org/10.1088/1361-6404/aa9584](https://doi.org/10.1088/1361-6404/aa9584)
    12. Olver, F. W. J., & Maximon, L. C. (2010). *NIST Handbook of Mathematical Functions*. Cambridge University Press. ISBN: [0521192250](https://www.worldcat.org/isbn/0521192250) ([9780521192255](https://www.worldcat.org/isbn/9780521192255))
    13. Cordeiro, M., Bezerra, I. P., & Vasconcelos, H. H. M. (2024). *Efficient computation of the wave function ψn(x) using Hermite coefficient matrix in Python*. In 7º Workshop Escola de Computação e Informação Quântica (7ª WECIQ) (pp. 56-60). CEFET/RJ.

    ## 🤝 Contributing

    Contributions, whether filing an issue, proposing improvements, or submitting a pull request, are welcome! Please feel free to explore, ask questions, and share your ideas.

    ## 📜 License

    This project is available under the *BSD 3-Clause License*. See the LICENSE file for more details.

    ## 📬 Contact

    If you have any questions or want to reach out to the team, please send me an email at [matheusgomescord@gmail.com](matheusgomescord@gmail.com).

"""



# BSD 3-Clause License
#
# Copyright (c) 2024, Matheus Gomes Cordeiro 
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
from sympy import symbols, diff, exp, Poly

import numpy as np
import numba as nb

# Adjusting numba settings to optimize parallel computations
nb.set_num_threads(nb.get_num_threads())

# Global variables for coefficient matrix and compilation status check
c_s_matrix = None

compilation_test = None

def hermite_sympy(n):
    """
    Compute the nth Hermite polynomial using symbolic differentiation.

    Args:
        n (int): Order of the Hermite polynomial.

    Returns:
        `sympy.Poly` : The nth Hermite polynomial as a sympy expression.

    Examples:
        >>> hermite_sympy(2)
        4*x**2 - 2

    References:
        1. Olver, F. W. J., & Maximon, L. C. (2010). NIST Handbook of Mathematical Functions. Cambridge University Press. https://search.worldcat.org/pt/title/502037224?oclcNum=502037224
        2. NIST Digital Library of Mathematical Functions. https://dlmf.nist.gov/, Release 1.0.28 of 2020-09-15.
        3. Sympy Documentation: https://docs.sympy.org/latest/modules/polys/index.html
    """

    x = symbols("x")
    return 1 if n == 0 else ((-1) ** n) * exp(x ** 2) * diff(exp(-x ** 2), x, n)


def create_normalized_hermite_coefficients_matrix(n_max):
    """
    Create a matrix of coefficients for normalized Hermite polynomials up to order `n_max`.

    Args:
        n_max (int): The maximum order of Hermite polynomials to compute.

    Returns:
        `np.ndarray` : A 2D numpy array containing the coefficients for the Hermite polynomials.

    Examples:
        >>> create_normalized_hermite_coefficients_matrix(3)
        array([[ 0.        ,  0.        ,  0.        ,  0.75112554],
               [ 0.        ,  0.        ,  1.06225193,  0.        ],
               [ 0.        ,  1.06225193,  0.        , -0.53112597],
               [ 0.86732507,  0.        , -1.30098761,  0.        ]])

    References:
        1. Olver, F. W. J., & Maximon, L. C. (2010). NIST Handbook of Mathematical Functions. Cambridge University Press.
        2. NIST Digital Library of Mathematical Functions. https://dlmf.nist.gov/, Release 1.0.28 of 2020-09-15.
        3. Sympy Documentation: https://docs.sympy.org/latest/modules/polys/index.html
    """

    x = symbols("x")
    C_s = np.zeros((n_max + 1, n_max + 1), dtype=np.float64)
    C_s[0, n_max] = 1

    for n in range(1, n_max + 1):
        c = Poly(hermite_sympy(n), x).all_coeffs()
        for index in range(n, -1, -1):
            C_s[n, (n_max + 1) - index - 1] = float(c[n - index])

    for i in range(n_max + 1):
        C_s[i] /=  (np.pi**0.50 * (2**i) * math.gamma(i+1))**0.5

    return C_s 


@nb.jit(nopython=True, looplift=True, nogil=True, boundscheck=False, cache=True)
def psi_n_single_fock_single_position(n, x, CS_matrix = True):
    """
    Compute the wavefunction for a real scalar `x` using a pre-computed matrix of normalized Hermite polynomial coefficients 
    until n=60 and then use the adapted recurrence relation for higher orders.

    Args:
        n (int): Quantum state number.
        x (float): Position at which to evaluate the wavefunction.
        CS_matrix (bool, optional): If True, use the optimized method for n <= 60, which relies on a pre-computed matrix 
                                        of coefficients for faster computation. For n > 60 or if False, use the general recursion 
                                        method. Defaults to True.

    Returns:
        `float` : The evaluated wavefunction.

    Examples:
        >>> psi_n_single_fock_single_position(0, 1.0)
        0.45558067201133257
        >>> psi_n_single_fock_single_position(61, 1.0)
        -0.2393049199171131

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """      
    
    if(n<=60 and CS_matrix):
        c_size = c_s_matrix.shape[0]
        n_coeffs = c_s_matrix[n]
        result = 0.0
        for i in range(c_size-n-1,c_size,2):
            result += n_coeffs[i]*(x**(c_size-i-1))

        return result * np.exp(-(x ** 2) / 2)
    else:
        r0 = 0.0
        r1 = (np.pi ** (-0.25)) * np.exp(-(x ** 2) / 2)
        r2 = 0.0

        for index in range(n):
            r2 = 2 * x * (r1 / np.sqrt(2 * (index + 1))) - np.sqrt(index / (index + 1)) * r0 
            r0 = r1
            r1 = r2

        return r1
    

@nb.jit(nopython=True, looplift=True, nogil=True, boundscheck=False, cache=True)
def psi_n_single_fock_single_position_complex(n, x, CS_matrix = True):
    """
    Compute the wavefunction for a complex scalar `x` using a pre-computed matrix of normalized Hermite polynomial coefficients 
    until n=60 and then use the adapted recurrence relation for higher orders.

    Args:
        n (int): Quantum state number.
        x (complex): Position at which to evaluate the wavefunction.
        CS_matrix (bool, optional): If True, use the optimized method for n <= 60, which relies on a pre-computed matrix 
                                     of coefficients for faster computation. For n > 60 or if False, use the general recursion 
                                     method. Defaults to True.

    Returns:
        `complex`: The evaluated wavefunction.

    Examples:
        >>> psi_n_single_fock_single_position_complex(0, 1.0 + 2.0j)
        (-1.4008797330262455 - 3.0609780602975003j)
        >>> psi_n_single_fock_single_position_complex(61, 1.0 + 2.0j)
        (-511062135.47555304 + 131445997.75753704j)

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """

    if(n<=60 and CS_matrix):
        c_size = c_s_matrix.shape[0]
        n_coeffs = c_s_matrix[n]
        result = 0.0 + 0.0j
        for i in range(c_size-n-1,c_size,2):
            result += n_coeffs[i]*(x**(c_size-i-1))

        return result * np.exp(-(x ** 2) / 2)
    else:
        r0 = 0.0 + 0.0j
        r1 = (np.pi ** (-0.25)) * np.exp(-(x ** 2) / 2)
        r2 = 0.0 + 0.0j

        for index in range(n):
            r2 = 2 * x * (r1 / np.sqrt(2 * (index + 1))) - np.sqrt(index / (index + 1)) * r0 
            r0 = r1
            r1 = r2

        return r1


@nb.jit(nopython=True, looplift=True,nogil=True, boundscheck=False, cache=True)
def psi_n_single_fock_multiple_position(n, x, CS_matrix = True):
    """
    Compute the wavefunction for a real vector `x` using a pre-computed matrix of normalized Hermite polynomial coefficients 
    until n=60 and x_size = 35. For higher orders, use the adapted recurrence relation.

    Args:
        n (int): Quantum state number.
        x (numpy.ndarray): Positions at which to evaluate the wavefunction.
        CS_matrix (bool, optional): If True, use the optimized method for n <= 60 and x_size <= 35, which relies on a pre-computed matrix 
                                     of coefficients for faster computation. For n > 60, or x_size > 35 or if False, use the general recursion 
                                     method. Defaults to True.

    Returns:
        `numpy.ndarray`: The evaluated wavefunction.

    Examples:
        >>> psi_n_single_fock_multiple_position(0, np.array([1.0, 2.0]))
        array([0.45558067, 0.10165379])
        >>> psi_n_single_fock_multiple_position(61, np.array([1.0, 2.0]))
        array([-0.23930492, -0.01677378])

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """

    x_size = x.shape[0]

    if(n<=60 and x_size<= 35 and CS_matrix):
        c_size = c_s_matrix.shape[0]
        n_coeffs = c_s_matrix[n]
        result = np.array([0.0] * (x_size))
        for j in range(x_size):
            for i in range(c_size-n-1,c_size,2):
                result[j] += n_coeffs[i]*(x[j]**(c_size-i-1))
            result[j] *= np.exp(-(x[j] ** 2) / 2)
        return result
    else:
        result = np.array([[0.0]*(x_size)]*(n+1))
        result[0] = (np.pi ** (-0.25))*np.exp(-(x ** 2) / 2) 

        for index in range(n):
            result[index+1]  = 2*x*(result[index]/np.sqrt(2*(index+1))) - np.sqrt(index/(index+1)) * result[index-1]
            
        return result[-1]
    


@nb.jit(nopython=True, looplift=True,nogil=True, boundscheck=False, cache=True)
def psi_n_single_fock_multiple_position_complex(n, x, CS_matrix = True):
    """
    Compute the wavefunction for a complex vector `x` using a pre-computed matrix of normalized Hermite polynomial coefficients 
    until n=60 and x_size = 35. For higher orders, use the adapted recurrence relation.

    Args:
        n (int): Quantum state number.
        x (numpy.ndarray): Positions at which to evaluate the wavefunction.
        CS_matrix (bool, optional): If True, use the optimized method for n <= 60 and x_size <= 35, which relies on a pre-computed matrix 
                                     of coefficients for faster computation. For n > 60, or x_size > 35 or if False, use the general recursion 
                                     method. Defaults to True.

    Returns:
        `numpy.ndarray`: The evaluated wavefunction.

    Examples:
        >>> psi_n_single_fock_multiple_position_complex(0, np.array([1.0 + 1.0j, 2.0 + 2.0j]))
        array([ 0.40583486-0.63205035j, -0.49096842+0.56845369j])
        >>> psi_n_single_fock_multiple_position_complex(61, np.array([1.0 + 1.0j, 2.0 + 2.0j]))
        array([-7.56548941e+03+9.21498621e+02j, -1.64189542e+08-3.70892077e+08j])

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """

    x_size = x.shape[0]

    if(n<=60 and x_size<= 35 and CS_matrix):
        c_size = c_s_matrix.shape[0]
        n_coeffs = c_s_matrix[n]
        result = np.array([0.0 + 0.0j] * (x_size))
        for j in range(x_size):
            for i in range(c_size-n-1,c_size,2):
                result[j] += n_coeffs[i]*(x[j]**(c_size-i-1))
            result[j] *= np.exp(-(x[j] ** 2) / 2)
        return result
    else:
        result = np.array([[0.0 + 0.0j]*(x_size)]*(n+1))
        result[0] = (np.pi ** (-0.25))*np.exp(-(x ** 2) / 2) 

        for index in range(n):
            result[index+1]  = 2*x*(result[index]/np.sqrt(2*(index+1))) - np.sqrt(index/(index+1)) * result[index-1]
            
        return result[-1]


@nb.jit(nopython=True, looplift=True,nogil=True, boundscheck=False, cache=True)
def psi_n_multiple_fock_single_position(n, x):
    """
    Compute the wavefunction for a real scalar `x` to all Fock states up to `n` using the recurrence relation.

    Args:
        n (int): Quantum state number.
        x (float): Position at which to evaluate the wavefunction.

    Returns:
        `numpy.ndarray`: The evaluated wavefunction.

    Examples:
        >>> psi_n_multiple_fock_single_position(1, 1.0)
        array([0.45558067, 0.64428837])

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """

    result = np.array([0.0] * (n+1))
    result[0] = (np.pi ** (-0.25))*np.exp(-(x ** 2) / 2) 

    for index in range(n):
        result[index+1]  = 2*x*(result[index]/np.sqrt(2*(index+1))) - np.sqrt(index/(index+1)) * result[index-1]
        
    return result


@nb.jit(nopython=True, looplift=True,nogil=True, boundscheck=False, cache=True)
def psi_n_multiple_fock_single_position_complex(n, x): 
    """
    Compute the wavefunction for a complex scalar `x` to all Fock states up to `n` using the recurrence relation.

    Args:
        n (int): Quantum state number.
        x (complex): Position at which to evaluate the wavefunction.

    Returns:
        `numpy.ndarray`: The evaluated wavefunction.

    Examples:
        >>> psi_n_multiple_fock_single_position_complex(1, 1.0 + 2.0j)
        array([-1.40087973-3.06097806j,  6.67661026-8.29116292j])

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """
    
    result = np.array([0.0 + 0.0j] * (n+1))
    result[0] = (np.pi ** (-0.25))*np.exp(-(x ** 2) / 2) 

    for index in range(n):
        result[index+1]  = 2*x*(result[index]/np.sqrt(2*(index+1))) - np.sqrt(index/(index+1)) * result[index-1]
        
    return result



@nb.jit(nopython=True, looplift=True,nogil=True, boundscheck=False, cache=True)
def psi_n_multiple_fock_multiple_position(n, x):
    """
    Compute the wavefunction for a real vector `x` to all Fock states up to `n` using the recurrence relation.

    Args:
        n (int): Quantum state number.
        x (numpy.ndarray): Positions at which to evaluate the wavefunction.

    Returns:
        `numpy.ndarray`: The evaluated wavefunction.

    Examples:
        >>> psi_n_multiple_fock_multiple_position(1, np.array([1.0, 2.0]))
        array([[0.45558067, 0.10165379],
               [0.64428837, 0.28752033]])

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """
    
    x_size = x.shape[0]
    result = np.array([[0.0]*(x_size)]*(n+1))
    result[0] = (np.pi ** (-0.25))*np.exp(-(x ** 2) / 2) 

    for index in range(n):
        result[index+1]  = 2*x*(result[index]/np.sqrt(2*(index+1))) - np.sqrt(index/(index+1)) * result[index-1]
        
    return result


@nb.jit(nopython=True, looplift=True,nogil=True, boundscheck=False, cache=True)
def psi_n_multiple_fock_multiple_position_complex(n, x):
    """
    Compute the wavefunction for a complex vector `x` to all Fock states up to `n` using the recurrence relation.

    Args:
        n (int): Quantum state number.
        x (numpy.ndarray): Positions at which to evaluate the wavefunction.

    Returns:
        `numpy.ndarray`: The evaluated wavefunction.

    Examples:
        >>> psi_n_multiple_fock_multiple_position_complex(1, np.array([1.0 + 1.0j, 2.0 + 2.0j]))
        array([[ 0.40583486-0.63205035j, -0.49096842+0.56845369j],
               [ 1.46779135-0.31991701j, -2.99649822+0.21916143j]])

    References:
        1. Pérez-Jordá, J. M. (2017). On the recursive solution of the quantum harmonic oscillator. *European Journal of Physics*, 
           39(1), 015402. doi: https://iopscience.iop.org/article/10.1088/1361-6404/aa9584
    """
    
    x_size = x.shape[0]
    result = np.array([[0.0 + 0.0j]*(x_size)]*(n+1))
    result[0] = (np.pi ** (-0.25))*np.exp(-(x ** 2) / 2) 

    for index in range(n):
        result[index+1]  = 2*x*(result[index]/np.sqrt(2*(index+1))) - np.sqrt(index/(index+1)) * result[index-1]
        
    return result
        
"""
Main execution block to initialize the coefficient matrix and test the wavefunction computation.

References
----------
- Python's official documentation offers insights into best practices for structuring and executing Python scripts:
  https://docs.python.org/3/tutorial/modules.html 
  .
"""

c_s_matrix = create_normalized_hermite_coefficients_matrix(60)

try:

    # Basic functionality test
    test_output_sfsp_CS_matrix_2 = psi_n_single_fock_single_position(2, 10.0)
    test_output_sfsp_2 = psi_n_single_fock_single_position(2, 10.0, CS_matrix=False)
    test_output_sfsp_CS_matrix_61 = psi_n_single_fock_single_position(61, 10.0)
    test_output_sfsp_61 = psi_n_single_fock_single_position(61, 10.0, CS_matrix=False)
    test_output_mfsp = psi_n_multiple_fock_single_position(2, 10.0)
    test_output_sfmp_CS_matrix_2 = psi_n_single_fock_multiple_position(2, np.array([10.0,4.5]))
    test_output_sfmp_2 = psi_n_single_fock_multiple_position(2, np.array([10.0,4.5]), CS_matrix=False)
    test_output_sfmp_CS_matrix_61 = psi_n_single_fock_multiple_position(61, np.array([10.0,4.5]))
    test_output_sfmp_61 = psi_n_single_fock_multiple_position(61, np.array([10.0,4.5]), CS_matrix=False)
    test_output_mfmp = psi_n_multiple_fock_multiple_position(2, np.array([10.0,4.5]))
    test_output_sfsp_c_CS_matrix_2 = psi_n_single_fock_single_position_complex(2, 10.0 + 0.0j)
    test_output_sfsp_c_2 = psi_n_single_fock_single_position_complex(2, 10.0 + 0.0j, CS_matrix=False)
    test_output_sfsp_c_CS_matrix_61 = psi_n_single_fock_single_position_complex(61, 10.0 + 0.0j)
    test_output_sfsp_c_61 = psi_n_single_fock_single_position_complex(61, 10.0 + 0.0j, CS_matrix=False)
    test_output_mfsp_c = psi_n_multiple_fock_single_position_complex(2, 10.0 + 0.0j)
    test_output_sfmp_c_CS_matrix_2 = psi_n_single_fock_multiple_position_complex(2, np.array([10.0 + 0.0j,4.5 + 0.0j]))
    test_output_sfmp_c_2 = psi_n_single_fock_multiple_position_complex(2, np.array([10.0 + 0.0j,4.5 + 0.0j]), CS_matrix=False)
    test_output_sfmp_c_CS_matrix_61 = psi_n_single_fock_multiple_position_complex(61, np.array([10.0 + 0.0j,4.5 + 0.0j]))
    test_output_sfmp_c_61 = psi_n_single_fock_multiple_position_complex(61, np.array([10.0 + 0.0j,4.5 + 0.0j]), CS_matrix=False)
    test_output_mfmp_c = psi_n_multiple_fock_multiple_position_complex(2, np.array([10.0 + 0.0j,4.5 + 0.0j]))
    compilation_test = True
    print(f"Functionality Test Passed: {compilation_test}")
except Exception as e:
    compilation_test = False
    print(f"Numba Compilation Error: {e}")



import numpy as np
import mpmath
import src.fast_wave.wavefunction_mpmath as wm

def test_wavefunction_computation():
    """
    Tests the basic functionality of all wavefunction_arb_prec functions.
    """

    # Testing basic functionality
    test_output_sfsp = wm.psi_n_single_fock_single_position(2, 10.0, 20)
    assert isinstance(test_output_sfsp, mpmath.ctx_mp_python.mpf)
    
    test_output_mfsp = wm.psi_n_multiple_fock_single_position(2, 10.0, 20)
    assert isinstance(test_output_mfsp , mpmath.matrices.matrices._matrix)
    
    test_output_sfmp = wm.psi_n_single_fock_multiple_position(2, np.array([10.0, 4.5]), 20)
    assert isinstance(test_output_sfmp, mpmath.matrices.matrices._matrix)
    
    test_output_mfmp = wm.psi_n_multiple_fock_multiple_position(2, np.array([10.0, 4.5]), 20)
    assert isinstance(test_output_mfmp, mpmath.matrices.matrices._matrix)
    
    test_output_sfsp_c = wm.psi_n_single_fock_single_position_complex(2, 10.0 + 0.0j, 20)
    assert isinstance(test_output_sfsp_c, mpmath.ctx_mp_python.mpc)
    
    test_output_mfsp_c = wm.psi_n_multiple_fock_single_position_complex(2, 10.0 + 0.0j, 20)
    assert isinstance(test_output_mfsp_c, mpmath.matrices.matrices._matrix)

    test_output_sfmp_c = wm.psi_n_single_fock_multiple_position_complex(2, np.array([10.0 + 0.0j, 4.5 + 0.0j]), 20)
    assert isinstance(test_output_sfmp_c, mpmath.matrices.matrices._matrix)
    
    test_output_mfmp_c = wm.psi_n_multiple_fock_multiple_position_complex(2, np.array([10.0 + 0.0j, 4.5 + 0.0j]), 20)
    assert isinstance(test_output_mfmp_c, mpmath.matrices.matrices._matrix)
    
    print("All functionality tests passed.")




import setuptools
from setuptools import find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

name = "fast_wave"
version = "1.6.0"
description = "Package for the calculation of the time-independent wavefunction." 
author_email = "matheusgomescord@gmail.com"
url = "https://github.com/pikachu123deimos/fast-wave" 

install_requires = [
    "llvmlite==0.42.0",
    "mpmath==1.3.0",
    "numba==0.59.1",
    "numpy==1.26.4",
    "sympy==1.12",
    "cython==3.0.10",
]

test_requires = [
    "pytest= ^7.1.2",
]

packages = find_packages(where='src')

package_data = {'fast_wave': ['*.pyd', '*.so']}

classifiers = [
    'Programming Language :: Python :: 3.11',  
    'Topic :: Scientific/Engineering :: Mathematics',  
    'License :: OSI Approved :: BSD License',  
    'Development Status :: 5 - Production/Stable', 
]

setuptools.setup(
    name=name,
    version=version,
    long_description=long_description,
    long_description_content_type="text/markdown",
    description=description,
    author=name,
    author_email=author_email,
    url=url,
    install_requires=install_requires,
    test_requires=test_requires,
    packages=packages,
    package_dir={'': 'src'},
    package_data=package_data,
    classifiers=classifiers
)
